//-----------------------------------------------------------------------------
// Init
//-----------------------------------------------------------------------------

const cluster = require('cluster');
const numCPUs = require('os').cpus().length;
const debug = require('debug')('cluster');

//-----------------------------------------------------------------------------
// Main
//-----------------------------------------------------------------------------

startCluster = function( startClusterCallback ){

  var setUpOneFork = function(){
    // cluster.settings = { stdio: [null, 'pipe', 'pipe']}
    var worker = cluster.fork();

    /// bind the messaging between each worker and the master
    worker.on('message', function(msg){
      debug('worker message: ', msg)
      // debug('namespace')

      //#set external kill switch
      //msg contains the processId
      //means the worker has started a sandbox
      //if time runs out, kill the process (cannot kill reliably within process)
      var timeoutID
      if(  !isNaN( msg ) ){
        timeoutID = setTimeout(function(){
           //kills itself
           process.kill( msg )
        }, 1000 );
      }

      if( msg === "clearTimeout" ){
        clearTimeout( timeoutID )
      }

      ///when at least one worker http server is listening.
      if ( msg === 'ready' && startClusterCallback ){
        // return that at least one server is running.
        startClusterCallback()
        //run once, empty the function
        startClusterCallback = function(){}
      }
    });

    worker.on('exit', function(code, signal){

      debug('worker '+ worker.process.pid +' died.');
      //start a new worker
      var newWorker = setUpOneFork() //cluster.fork();
      debug('worker '+ newWorker.process.pid + ' born.');
    });

    return worker
  }


  cluster.setupMaster({
    exec: __dirname + '/http-worker.js'
  });

  if (cluster.isMaster) {
    // CREATE A CLUSTER OF FORKED SCRIPTS, ONE PER CPU
    debug(`Master ${process.pid} is running`);

    for ( var i = 0; i < numCPUs; i++ ) {
      setUpOneFork()
    }
  }

  return process.pid
}


//-----------------------------------------------------------------------------
// EXPORTS
//-----------------------------------------------------------------------------


module.exports = startCluster;
