//-----------------------------------------------------------------------------
// Init
//-----------------------------------------------------------------------------

// const requestlib = require("request");
const intercept = require("intercept-stdout");
const vm = require('vm');
const acorn = require('acorn');
const walk = require( 'estree-walker' ).walk;
const _ = require('lodash')
const debug = require('debug')('sandbox');


//-----------------------------------------------------------------------------
// Main
//-----------------------------------------------------------------------------
const hijackReturnStatement = function( code ){
  var timeA = Date.now()

  var ast = acorn.parse(code);
  var returnInstances

  walk( ast, {
    enter: function ( node, parent ) {
      if( node.type == 'VariableDeclarator' && node.id.name == "test"  ){
        returnInstances = node.init.body.body
        this.skip()//stop recursive walk when this is found.
      }
    },
  });

  var returnDoc = _.filter( returnInstances, function ( each ){
    return each.type === "ReturnStatement"
  })

  // debug('mainFunction', JSON.stringify( returnDoc, null, 4 ) );
  // debug('return argument', code.slice( returnDoc[0].argument.start, returnDoc[0].argument.end ) );
  var returnArg = code.slice( returnDoc[0].argument.start, returnDoc[0].argument.end )
  newCode = code.slice( 0 ,returnDoc[0].start ) + 'done('+ returnArg + ')' + code.slice( returnDoc[0].end )
  // debug('mainFunction', newCode );


  debug('time:', Date.now() - timeA );
  return newCode
}

const sandbox = function(script, whiteListedLibraries, defaultContext, callback ){
  // contextify needs to be run new everytime.

  var Contextify = new require('contextify');

////////////// PREP ARGS FOR VM //////////////////////
  timeA =  Date.now()
  self = this
  timeoutValue = 5000 //millisecond timemout on all sandbox ops
  var sandboxOutput = {}
  var exitCode = 0
  var sandboxLogs = []
  var sandboxError

  if ( !defaultContext ) {
    var defaultContext = {}
  }
    //standard utils
  defaultContext.setTimeout = setTimeout
  defaultContext.console = console
  defaultContext.Buffer = Buffer
  defaultContext.JSON = JSON


  //// TODO: GROSS, this function should be defined in wireline, not in module
  // defaultContext.httpRequest = function( requestObj ) {
  //
  //     return new Promise(function (resolve, reject) {
  //       requestlib( requestObj , function (err, IncomingMessage, body) {
  //         if (err) {
  //             // console.log('httpRequest err', err);
  //             return reject(err);
  //         }
  //         // console.log('IncomingMessage.statusCode', IncomingMessage.statusCode);
  //         // if (IncomingMessage.statusCode !== 200) {
  //         //     err = new Error("Unexpected status code: " + res.statusCode);
  //         //     err.res = res;
  //         //     return reject(err);
  //         // }
  //         // console.log('httpRequest res', res);
  //         resolve(body);
  //       });
  //     });
  //
  //   }

    if ( whiteListedLibraries ) {
      for (var i = 0, len = whiteListedLibraries.length; i < len; i++) {
        // import each library with require as necessary.
        defaultContext[ whiteListedLibraries[i].namespace ]   =  require( whiteListedLibraries[i].npm );
      }
    }

    try{
      Contextify = Contextify( defaultContext )

    }catch(e){
      debug('sandbox.js - contextify error', e);
    }


  /// TODO: GROSS I don't like this. How can we tell between sync and asyc scripts

  if ( script.indexOf( 'done(') > -1 ){
    var wrappedCode = "(function(done) {" + script + "; })"
  }else{
    var wrappedCode = "(function(done) {" + script + "; done()})"
  }

  try {
    debug('sandbox.js - running');
    //ASYNC
    timeA =  Date.now()

    //intercept logs
    //https://www.npmjs.com/package/intercept-stdout
    interceptedLogs = [];
    //this initializes the intercept and calling it will end the intercept
    var endIntercept = intercept(function(txt) {
        interceptedLogs.push(txt) //+= txt;
    });


    //// TODO: NODE 4.5.0 implementation
    // var newContext = vm.createContext( defaultContext )
    //
    // var sandbox2 = vm.runInContext( wrappedCode, newContext )(function(){
    //   // console.log('time Contextify run : ', Date.now() - timeA );
    //
    //    //the sandbox returns a string. Need to parse again
    //    var sandboxResult = newContext.RESULT
    //
    //    //message the parent sandbox.js with the result
    //
    //   //  TODO: should we handle an undefined?
    //   //  if ( sandboxResult === undefined ){
    //   //    exitCode = 1
    //   //    sandboxResult = 'undefined'
    //   //  }
    //
    //    sandboxOutput = { status: 'COMPLETE',result: sandboxResult }
    //
    //    exitCode = 0
    //    //kills the sandbox after the timeout
    //    Contextify.dispose();
    //
    //    //stop intercepting console logs
    //    endIntercept();
    //    sandboxOutput.console = interceptedLogs
    //
    //   //  console.log("This text is not being captured");
    //    console.log('captured_text',  interceptedLogs );
    //
    //
    //    if ( callback && exitCode === 0  ){
    //      callback(null, sandboxOutput)
    //    }
    //
    // })
    // console.log('newContext ', newContext.RES );

//////////////// NODE 0.10 implementation: /////////////////
    Contextify.run( wrappedCode )(function( callbackResult ) {
      // console.log('result in callback', callbackResult);
      // console.log('process outside', process.pid);

      // console.log('Contextify obj', Contextify);
      // console.log('time Contextify run : ', Date.now() - timeA );

    var sandboxResult

     if( callbackResult ){
      sandboxResult = callbackResult
    }else{
      sandboxResult = Contextify.RESULT
    }
     //the sandbox returns a string. Need to parse again

     //message the parent sandbox.js with the result

    //  TODO: should we handle an undefined?
    //  if ( sandboxResult === undefined ){
    //    exitCode = 1
    //    sandboxResult = 'undefined'
    //  }

     sandboxOutput = { status: 'COMPLETE',result: sandboxResult }

     exitCode = 0
     //kills the sandbox after the timeout
     Contextify.dispose();

     //stop intercepting console logs
     endIntercept();
     sandboxOutput.console = interceptedLogs

    //  console.log("This text is not being captured");
     debug('sandbox.js - captured_text',  interceptedLogs );


     if ( callback && exitCode === 0  ){
       callback(null, sandboxOutput)
     }

    //  exits with error or kill signal
     if ( callback && (exitCode != 0 ) ){
       callback( sandboxOutput , null)
     }

     debug('sandbox2.js - Async Complete')

    });
    //////////////////////// END NODE 0.10 implementation ////////////////////////
  }catch (e){
   // send message back to parent sandbox.js  with handled  exceptions from the sandbox
    //endIntercept may not have ended
    endIntercept()
    sandboxOutput = { status: 'ERROR', result: e.name +" : " +  e.message, console:  interceptedLogs}
    debug('sandbox.js - error: ', sandboxOutput);
    // process.send('sandbox.js - error: ', sandboxOutput);

    exitCode = 1
    if ( callback && (exitCode != 0 ) ){
      callback( sandboxOutput , null)
    }
  }
}


//-----------------------------------------------------------------------------
// EXPORTS
//-----------------------------------------------------------------------------

module.exports = sandbox;
